// Copyright (C) 2012, Clément Pit--Claudel (http://pit-claudel.fr/clement/blog)
// 
// Permission is hereby granted, free of charge, to any person obtaining a copy of 
// this software and associated documentation files (the "Software"), to deal in 
// the Software without restriction, including without limitation the rights to 
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
// the Software, and to permit persons to whom the Software is furnished to do so, 
// subject to the following conditions:
// 
// The above copyright notice and this permission notice shall be included in all 
// copies or substantial portions of the Software.
// 
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR 
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
// FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR 
// COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER 
// IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
// CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

#include <cstdio>
#include <cstdlib>
#include <vector>
#include <string>
#define NOMINMAX
#include <windows.h>
#include <iostream>

const int count = 100*1000;
const int repeat = 20;
const int max_length = 500;

inline double min(double a, double b) {return (a < b) ? a : b;}
std::string allowed_chars = std::string("abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ");

std::vector<std::string> strings1 = std::vector<std::string>(count);
std::vector<std::string> strings2 = std::vector<std::string>(count);

double elapsed_ms(LARGE_INTEGER start, LARGE_INTEGER end) {
	LARGE_INTEGER frequency;
	QueryPerformanceFrequency(&frequency);
	return (end.QuadPart - start.QuadPart) * 1000 / (double) frequency.QuadPart;
}

int mystrcmp(char* a, char* b) {
    int count = 0;
    char* aa = a, *bb = b;
    while (*a == *b && *a != '\0' && *b != '\0') {
        //printf("%d %s %s\n", count, a, b);
        a++; b++;
        count++;
    }
    if (count > 15)
        printf("%s %s!!!\n", aa, bb);
    return ((*a == 0 || *b == 0) ? 0 : 1);
}

void compare_cstrings(int count, int length, int iterations) {
	std::vector<char*> shortstrings1 = std::vector<char*>(count);
	std::vector<char*> shortstrings2 = std::vector<char*>(count);

	LARGE_INTEGER start, end;

	for (int str_id = 0; str_id < count; str_id++) {
		shortstrings1[str_id] = new char[length + 1];
		shortstrings2[str_id] = new char[length + 1];
        strings1[str_id].copy(shortstrings1[str_id], length, 0);
        strings2[str_id].copy(shortstrings2[str_id], length, 0);
        shortstrings1[str_id][length] = shortstrings2[str_id][length] = '\0';
	}

	double best;
	int matches, non_matches;

	for (int iteration_id = 0; iteration_id < iterations; iteration_id++) {
		matches = 0, non_matches = 0;

		QueryPerformanceCounter(&start);
		for (int str_id = 0; str_id < count; str_id++) {
			if (strcmp(shortstrings1[std::rand()%count], shortstrings2[std::rand()%count]) == 0)
				matches++;
            else
                non_matches++;
        }
		QueryPerformanceCounter(&end);

		double time = elapsed_ms(start, end);
		if (iteration_id == 0)
			best = time;
		else
			best = min(best, time);
	}

	for (int str_id = 0; str_id < count; str_id++) {
		delete[] shortstrings1[str_id];
		delete[] shortstrings2[str_id];
	}

	printf("C-Strings\t%d\t%d\t%f\tms\t(%d matches, %d non-matches)\n", count, length, best, matches, non_matches);
}

void compare_strings(int count, int length, int iterations) {
	std::vector<std::string> shortstrings1 = std::vector<std::string>(count);
	std::vector<std::string> shortstrings2 = std::vector<std::string>(count);

	LARGE_INTEGER start, end;

	for (int str_id = 0; str_id < count; str_id++) {
		shortstrings1[str_id] = strings1[str_id].substr(0, length);
		shortstrings2[str_id] = strings2[str_id].substr(0, length);
	}

	double best;
	int matches, non_matches;

	for (int iteration_id = 0; iteration_id < iterations; iteration_id++) {
		matches = 0, non_matches = 0;

		QueryPerformanceCounter(&start);
		for (int str_id = 0; str_id < count; str_id++) {
			if (shortstrings1[str_id] == shortstrings2[str_id])
				matches++;
			else
				non_matches++;
		}
		QueryPerformanceCounter(&end);

		double time = elapsed_ms(start, end);
		if (iteration_id == 0)
			best = time;
		else
			best = min(best, time);
	}

	printf("  Strings\t%d\t%d\t%f\tms\t(%d matches, %d non-matches)\n", count, length, best, matches, non_matches);
}

void compare_ints(int count, int iterations) {
	std::vector<int> ints1 = std::vector<int>(count);
	std::vector<int> ints2 = std::vector<int>(count);

	LARGE_INTEGER start, end;

	for (int int_id = 0; int_id < count; int_id++) {
		ints1[int_id] = std::rand();
		ints2[int_id] = std::rand();
	}

	double best;
	int matches, non_matches;

	for (int iteration_id = 0; iteration_id < iterations; iteration_id++) {
        matches = 0, non_matches = 0;

		QueryPerformanceCounter(&start);
		for (int int_id = 0; int_id < count; int_id++) {
			if (ints1[int_id] == ints2[int_id])
				matches++;
			else
				non_matches++;
		}
		QueryPerformanceCounter(&end);

		double time = elapsed_ms(start, end);
		if (iteration_id == 0)
			best = time;
		else
			best = min(best, time);
	}

	printf("Ints\t%d\t%f\tms\t(%d matches, %d non-matches)\n", count, best, matches, non_matches);
}

int main(int argc, char** argv) {
	srand(0);
	//int increment = 1000*1000, max = 1000*1000;

	for (int str_id = 0; str_id < count; str_id++) {
        std::string str1 = std::string(max_length, ' ');
        std::string str2 = std::string(max_length, ' ');

		for (int char_id = 0; char_id < max_length; char_id++) {
			str1[char_id] = allowed_chars[std::rand() % allowed_chars.length()];
			str2[char_id] = allowed_chars[std::rand() % allowed_chars.length()];
		}

		strings1[str_id] = str1;
		strings2[str_id] = str2;
	}

	//for (int count = increment; count <= max; count += increment)
	for (int length = 1; length <= max_length; length += 1)
		compare_cstrings(count, length, repeat);

	for (int length = 1; length <= max_length; length += 1)
		compare_strings(count, length, repeat);

	//for (int count = increment; count <= max; count += increment)
	compare_ints(count, repeat);

	scanf("%d", NULL);
}
